"""Script to generate Z-Wave device data."""
import json
import logging
import os
import re
import shutil
import tarfile
import threading
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Generator

import aiohttp
import async_timeout

_LOGGER = logging.getLogger(__name__)
UPDATED_AT = "updated_at"

TMP_DIR = os.path.join(os.path.dirname(__file__), "./.tmp/")
REPO_URL = "https://api.github.com/repos/zwave-js/node-zwave-js"
ZJS_TAR_FILE = os.path.join(TMP_DIR, "zjs.tar.gz")
ZJS_TAR_CONFIG_BASE = "/packages/config/config/"
ZJS_DEVICE_CONFIG_DB_FILE = os.path.join(
    os.path.dirname(__file__), "zjs_device_config_db.json"
)

__MUTEX = threading.Lock()


def get_zwave_device_info(
    manufacturer_id: int, product_type: int, product_id: int
) -> dict:
    """Lookup the Z-Wave device based on the manufacturer id, product type and product id."""
    key = f"0x{manufacturer_id:04x}:0x{product_type:04x}:0x{product_id:04x}"
    return _load_db_from_file().get(key, {})


def _device_config_db_file_exists() -> bool:
    """Return True if the device config db file exists."""
    return (
        os.path.isfile(ZJS_DEVICE_CONFIG_DB_FILE)
        and os.path.getsize(ZJS_DEVICE_CONFIG_DB_FILE) > 0
    )


def _load_db_from_file() -> dict:
    """Load the Z-Wave JS device config from the saved JSON file."""
    data = {}
    if _device_config_db_file_exists():
        with open(ZJS_DEVICE_CONFIG_DB_FILE) as f:
            data = json.load(f)
    return data


async def download_zjs_device_config_db() -> dict:
    """Download the Z-Wave JS device config database."""
    with __MUTEX:
        if await _is_new_version_available():
            start_date = datetime.now(timezone.utc).replace(tzinfo=None).isoformat()
            _LOGGER.debug("Beginning download process")
            _clean_temp_directory(create=True)
            await _download_zjs_tarfile()
            _extract_zjs_config_files()
            data = _create_db_from_zjs_config_files(updated_at=start_date)
            _clean_temp_directory()
            return data
        else:
            _LOGGER.debug("An updated file was not created")
            return _load_db_from_file()


async def _is_new_version_available() -> bool:
    """Return `True` if a newer archive of the repo at http://github.com/zwave-js/node-zwave-js is available."""
    file_updated_at = _load_db_from_file().get(UPDATED_AT)
    if file_updated_at is None:
        _LOGGER.debug("File has not yet been created")
        return True
    file_updated_at = datetime.fromisoformat(file_updated_at)

    _LOGGER.debug("Retrieving last updated date from %s", REPO_URL)
    async with aiohttp.ClientSession() as session:
        async with async_timeout.timeout(10):
            async with session.get(REPO_URL) as response:
                if response.status != 200:
                    _LOGGER.debug("Unable to check last updated date from %s", REPO_URL)
                    return False
                updated_at = datetime.fromisoformat(
                    (await response.json()).get(UPDATED_AT).replace("Z", "")
                )
                _LOGGER.debug(
                    "Repo was last updated at %s, while saved file was last updated at %s",
                    updated_at,
                    file_updated_at,
                )
                return file_updated_at < updated_at


def _clean_temp_directory(create: bool = False) -> None:
    """Ensure the temp directory is empty and create it if specified."""
    if os.path.exists(TMP_DIR):
        _LOGGER.debug("Removing temp directory")
        shutil.rmtree(TMP_DIR)

    if create:
        _LOGGER.debug("Creating temp directory")
        os.mkdir(TMP_DIR)


async def _download_zjs_tarfile() -> None:
    """Download the Z-Wave JS tarfile from http://github.com/zwave-js/node-zwave-js."""
    download_url = f"{REPO_URL}/tarball"
    _LOGGER.debug("Downloading tarfile from %s", download_url)
    async with aiohttp.ClientSession() as session:
        async with async_timeout.timeout(120):
            async with session.get(download_url) as response:
                with open(ZJS_TAR_FILE, "wb") as file:
                    async for data in response.content.iter_chunked(1024):
                        file.write(data)


def _extract_zjs_config_files() -> None:
    """Extract the Z-Wave JSON config files."""

    def members(tf: tarfile.TarFile) -> Generator[tarfile.TarInfo, Any, Any]:
        base_path = f"{tf.firstmember.path}{ZJS_TAR_CONFIG_BASE}"
        manufacturers_path = f"{base_path}manufacturers.json"
        devices_path = f"{base_path}devices/"
        base_length = len(base_path)
        for member in tf.getmembers():
            if member.path.startswith(manufacturers_path) or (
                member.path.startswith(devices_path)
                and member.path.endswith(".json")
                and "/templates/" not in member.path
            ):
                member.path = member.path[base_length:]
                yield member

    _LOGGER.debug("Extracting config files from download")
    with tarfile.open(ZJS_TAR_FILE) as tar:
        tar.extractall(members=members(tar), path=TMP_DIR)


def _create_db_from_zjs_config_files(updated_at: str) -> dict:
    """Parse the Z-Wave JSON config files and create a consolidated device db."""
    _LOGGER.debug("Parsing extracted config files")
    json_files = Path(os.path.join(TMP_DIR, "devices")).glob("**/*.json")

    device_db = {}

    for file in json_files:
        data = {}
        try:
            with open(file) as json_file:
                json_string = "".join(
                    re.sub("((^|\\s+)//.*)|(/\\*.*\\*/)", "", line)
                    for line in json_file.readlines()
                )
                data = json.loads(json_string)
        except:  # noqa: E722
            print("oops, couldn't parse file %s", file)

        manufacturer_id = data.get("manufacturerId")
        manufacturer = data.get("manufacturer")
        label = data.get("label")
        description = data.get("description")

        for device in data.get("devices", []):
            product_type = device.get("productType")
            product_id = device.get("productId")
            device_db[f"{manufacturer_id}:{product_type}:{product_id}"] = {
                "manufacturer": manufacturer,
                "label": label,
                "description": description,
            }

    if device_db == {}:
        _LOGGER.error("Unable to create consolidated device db")
    else:
        _LOGGER.debug("Creating consolidated device db")
        device_db.update({UPDATED_AT: updated_at})
        with open(ZJS_DEVICE_CONFIG_DB_FILE, "w") as device_file:
            device_file.write(json.dumps(device_db, sort_keys=True, indent=2))

    return device_db



"""Module that implements the VivintSkyApi class."""
from __future__ import annotations

import json
import logging
import ssl
from datetime import datetime
from types import MethodType
from typing import Any

import aiohttp
import certifi
from aiohttp import ClientResponseError

from .const import (
    AuthenticationResponse,
    MfaVerificationResponse,
    SwitchAttribute,
    VivintDeviceAttribute,
)
from .enums import ArmedState, GarageDoorState, ZoneBypass
from .exceptions import (
    VivintSkyApiAuthenticationError,
    VivintSkyApiError,
    VivintSkyApiMfaRequiredError,
)

_LOGGER = logging.getLogger(__name__)

VIVINT_API_ENDPOINT = "https://www.vivintsky.com/api"
VIVINT_MFA_ENDPOINT = (
    "https://www.vivintsky.com/platform-user-api/v0/platformusers/2fa/validate"
)


class VivintSkyApi:
    """Class to communicate with the VivintSky API."""

    def __init__(
        self,
        username: str,
        password: str,
        persist_session: bool = False,
        client_session: aiohttp.ClientSession | None = None,
    ):
        """Initialize the VivintSky API."""
        self.__username = username
        self.__password = password
        self.__persist_session = persist_session
        self.__client_session = client_session or self.__get_new_client_session()
        self.__has_custom_client_session = client_session is not None
        self.__mfa_pending = False

    def is_session_valid(self) -> dict:
        """Return the state of the current session."""
        cookie = self.__client_session.cookie_jar._cookies["www.vivintsky.com"].get("s")
        if not cookie:
            return False
        cookie_expiration = datetime.strptime(
            cookie.get("expires"), "%a, %d %b %Y %H:%M:%S %Z"
        )
        return True if cookie_expiration > datetime.utcnow() else False

    async def connect(self) -> dict:
        """Connect to VivintSky Cloud Service."""
        if self.__has_custom_client_session and self.is_session_valid():
            authuser_data = await self.get_authuser_data()
        else:
            authuser_data = await self.__get_vivintsky_session(
                self.__username, self.__password, self.__persist_session
            )
        if not authuser_data:
            raise VivintSkyApiAuthenticationError("Unable to login to Vivint")
        return authuser_data

    async def disconnect(self) -> None:
        """Disconnect from VivintSky Cloud Service."""
        if not self.__has_custom_client_session:
            await self.__client_session.close()

    async def verify_mfa(self, code: str) -> None:
        """Verify multi-factor authentication code."""
        resp = await self.__post(
            VIVINT_MFA_ENDPOINT,
            data=json.dumps({"code": code}),
        )
        if resp is not None:
            self.__mfa_pending = False

    async def get_authuser_data(self) -> dict:
        """
        Get the authuser data.

        Poll the Vivint authuser API endpoint resource to gather user-related data including enumeration of the systems
        that user has access to.
        """
        resp = await self.__get("authuser")
        if resp:
            return resp
        else:
            raise VivintSkyApiAuthenticationError("Missing auth user data")

    async def get_panel_credentials(self, panel_id: int) -> dict:
        """Get the panel credentials."""
        resp = await self.__get(f"panel-login/{panel_id}")
        if resp:
            return resp
        else:
            raise VivintSkyApiAuthenticationError(
                "Unable to retrieve panel credentials."
            )

    async def get_system_data(self, panel_id: int) -> dict:
        """Get the raw data for a system."""
        resp = await self.__get(
            f"systems/{panel_id}",
            headers={"Accept-Encoding": "application/json"},
            params={"includerules": "false"},
        )
        if resp:
            return resp
        else:
            raise VivintSkyApiError("Unable to retrieve system data")

    async def get_device_data(self, panel_id: int, device_id: int) -> dict:
        """Get the raw data for a device."""
        resp = await self.__get(
            f"system/{panel_id}/device/{device_id}",
            headers={"Accept-Encoding": "application/json"},
        )
        if resp:
            return resp
        else:
            raise VivintSkyApiError("Unable to retrieve device data")

    async def set_alarm_state(
        self, panel_id: int, partition_id: int, state: bool
    ) -> aiohttp.ClientResponse:
        """Set the alarm state."""
        resp = await self.__put(
            f"{panel_id}/{partition_id}/armedstates",
            headers={"Content-Type": "application/json;charset=UTF-8"},
            data=json.dumps(
                {
                    "system": panel_id,
                    "partitionId": partition_id,
                    "armState": state,
                    "forceArm": False,
                }
            ).encode("utf-8"),
        )
        if resp is None:
            _LOGGER.error(
                "Failed to set state to %s for panel %s",
                ArmedState(state).name,
                self.id,
            )
            raise VivintSkyApiError("Failed to set alarm state")

    async def set_garage_door_state(
        self, panel_id: int, partition_id: int, device_id: int, state: int
    ) -> None:
        """Open/Close garage door."""
        resp = await self.__put(
            f"{panel_id}/{partition_id}/door/{device_id}",
            headers={
                "Content-Type": "application/json;charset=utf-8",
            },
            data=json.dumps(
                {
                    VivintDeviceAttribute.STATE: state,
                    VivintDeviceAttribute.ID: device_id,
                }
            ).encode("utf-8"),
        )
        if resp is None:
            _LOGGER.debug(
                "Failed to set state to %s for garage door %s @ %s:%s",
                GarageDoorState(state).name,
                device_id,
                panel_id,
                partition_id,
            )
            raise VivintSkyApiError("Failed to set garage door state")

    async def set_lock_state(
        self, panel_id: int, partition_id: int, device_id: int, locked: bool
    ) -> None:
        """Lock/Unlock door lock."""
        resp = await self.__put(
            f"{panel_id}/{partition_id}/locks/{device_id}",
            headers={
                "Content-Type": "application/json;charset=utf-8",
            },
            data=json.dumps(
                {
                    VivintDeviceAttribute.STATE: locked,
                    VivintDeviceAttribute.ID: device_id,
                }
            ).encode("utf-8"),
        )
        if resp is None:
            _LOGGER.debug(
                "Failed to set state to %s for lock %s @ %s:%s",
                locked,
                device_id,
                panel_id,
                partition_id,
            )
            raise VivintSkyApiError("Failed to set lock state")

    async def set_sensor_state(
        self, panel_id: int, partition_id: int, device_id: int, bypass: bool
    ) -> None:
        """Bypass/unbypass a sensor."""
        resp = await self.__put(
            f"{panel_id}/{partition_id}/sensors/{device_id}",
            headers={
                "Content-Type": "application/json;charset=utf-8",
            },
            data=json.dumps(
                {
                    VivintDeviceAttribute.BYPASSED: ZoneBypass.MANUALLY_BYPASSED
                    if bypass
                    else ZoneBypass.UNBYPASSED,
                    VivintDeviceAttribute.ID: device_id,
                }
            ).encode("utf-8"),
        )
        if resp is None:
            _LOGGER.debug(
                "Failed to set state to %s for sensor %s @ %s:%s",
                "bypassed" if bypass else "unbypassed",
                device_id,
                panel_id,
                partition_id,
            )
            raise VivintSkyApiError("Failed to set sensor state")

    async def set_switch_state(
        self,
        panel_id: int,
        partition_id: int,
        device_id: int,
        on: bool | None = None,
        level: int | None = None,
    ) -> None:
        """Set switch state."""
        # validate input
        if on is None and level is None:
            raise VivintSkyApiError('Either "on" or "level" must be provided.')
        elif level and (0 > level or level > 100):
            raise VivintSkyApiError('The value for "level" must be between 0 and 100.')

        resp = await self.__put(
            f"{panel_id}/{partition_id}/switches/{device_id}",
            headers={
                "Content-Type": "application/json;charset=utf-8",
            },
            data=json.dumps(
                {
                    SwitchAttribute.ID: device_id,
                    **(
                        {SwitchAttribute.STATE: on}
                        if level is None
                        else {SwitchAttribute.VALUE: level}
                    ),
                }
            ).encode("utf-8"),
        )
        if resp is None:
            [attribute, value] = ["on", on] if level is None else ["level", level]
            _LOGGER.debug(
                "Failed to set %s to %s for switch %s @ %s:%s",
                attribute,
                value,
                device_id,
                panel_id,
                partition_id,
            )
            raise VivintSkyApiError("Failed to set switch state")

    async def set_thermostat_state(
        self, panel_id: int, partition_id: int, device_id: int, **kwargs
    ) -> None:
        """Set thermostat state."""
        resp = await self.__put(
            f"{panel_id}/{partition_id}/thermostats/{device_id}",
            headers={
                "Content-Type": "application/json;charset=utf-8",
            },
            data=json.dumps(kwargs).encode("utf-8"),
        )
        if resp is None:
            _LOGGER.debug(
                "Failed to set state to %s for thermostat %s @ %s:%s",
                kwargs,
                device_id,
                panel_id,
                partition_id,
            )
            raise VivintSkyApiError("Failed to set thermostat state")

    async def request_camera_thumbnail(
        self, panel_id: int, partition_id: int, device_id: int
    ) -> None:
        """Request the camera thumbnail."""
        try:
            return await self.__get(
                f"{panel_id}/{partition_id}/{device_id}/request-camera-thumbnail",
            )
        except ClientResponseError as resp:
            if resp.status < 200 or resp.status > 299:
                _LOGGER.error(
                    "Failed to request thumbnail for camera %s @ %s:%s with status code %s",
                    device_id,
                    panel_id,
                    partition_id,
                    resp.status,
                )

    async def get_camera_thumbnail_url(
        self,
        panel_id: int,
        partition_id: int,
        device_id: int,
        thumbnail_timestamp: datetime,
    ) -> str:
        """Get the camera thumbnail url."""
        try:
            resp = await self.__get(
                f"{panel_id}/{partition_id}/{device_id}/camera-thumbnail",
                params={"time": thumbnail_timestamp},
                allow_redirects=False,
            )
            return resp.get("location")
        except ClientResponseError as resp:
            if resp.status != 302:
                _LOGGER.debug(
                    "Failed to get thumbnail for camera %s @ %s:%s with status code %s",
                    device_id,
                    panel_id,
                    partition_id,
                    resp.status,
                )

    def __get_new_client_session(self) -> aiohttp.ClientSession:
        """Create a new aiohttp.ClientSession object."""
        ssl_context = ssl.create_default_context(
            purpose=ssl.Purpose.SERVER_AUTH, cafile=certifi.where()
        )
        connector = aiohttp.TCPConnector(enable_cleanup_closed=True, ssl=ssl_context)

        return aiohttp.ClientSession(connector=connector)

    async def __get_vivintsky_session(
        self, username: str, password: str, persist_session: bool = False
    ) -> dict:
        """Login into the Vivint Sky platform with the given username, password and, optionally, persist the session.

        Returns auth user data if successful.
        """
        return await self.__post(
            "login",
            data=json.dumps(
                {
                    "username": username,
                    "password": password,
                    "persist_session": persist_session,
                }
            ).encode("utf-8"),
        )

    async def __get(
        self,
        path: str,
        headers: dict[str, Any] = None,
        params: dict[str, Any] = None,
        allow_redirects: bool = None,
    ) -> dict | None:
        """Perform a get request."""
        return await self.__call(
            self.__client_session.get,
            path,
            headers=headers,
            params=params,
            allow_redirects=allow_redirects,
        )

    async def __post(
        self,
        path: str,
        data: bytes = None,
    ) -> dict | None:
        """Perform a post request."""
        return await self.__call(self.__client_session.post, path, data=data)

    async def __put(
        self,
        path: str,
        headers: dict[str, Any] = None,
        data: bytes = None,
    ) -> dict | None:
        """Perform a put request."""
        return await self.__call(
            self.__client_session.put, path, headers=headers, data=data
        )

    async def __call(
        self,
        method: MethodType,
        path: str,
        headers: dict[str, Any] = None,
        params: dict[str, Any] = None,
        data: bytes = None,
        allow_redirects: bool = None,
    ) -> dict | None:
        """Perform a request with supplied parameters and reauthenticate if necessary."""
        if path != "login" and not self.is_session_valid():
            await self.connect()

        if self.__client_session.closed:
            raise VivintSkyApiError("The client session has been closed")

        is_mfa_request = path == VIVINT_MFA_ENDPOINT

        if self.__mfa_pending and not is_mfa_request:
            raise VivintSkyApiMfaRequiredError(AuthenticationResponse.MFA_REQUIRED)

        resp = await method(
            path if is_mfa_request else f"{VIVINT_API_ENDPOINT}/{path}",
            headers=headers,
            params=params,
            data=data,
            allow_redirects=allow_redirects,
        )
        async with resp:
            data: dict = await resp.json(encoding="utf-8")
            if resp.status == 200:
                return data
            elif resp.status == 302:
                return {"location": resp.headers.get("Location")}
            elif resp.status == 401:
                message = (
                    data.get(MfaVerificationResponse.MESSAGE)
                    if is_mfa_request
                    else data.get(AuthenticationResponse.MESSAGE)
                )
                if message == AuthenticationResponse.MFA_REQUIRED:
                    self.__mfa_pending = True
                    raise VivintSkyApiMfaRequiredError(message)
                raise VivintSkyApiAuthenticationError(message)
            else:
                resp.raise_for_status()
                return None
